#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>


int main(int argc, char **argv) {
  struct sockaddr_in server;     // struct for server address
  struct sockaddr_in response;   // address of server from received packet
  int port;                      // server port number
  struct hostent *hp;            // for gethostbyname
  int sock;                      // TCP socket number
  char buf[1024];                // buffer for transmit and receive
  int len;                       // length of received block
  int rc;                        // result code
  int respsize;                  // size of response struct
  
  // Check to make sure the user passed the right number of arguments
  if (argc < 2) {
    printf("Usage: EchoClientUDP host [port]\n");
    exit(1);
  }

  // Set the server address from argument 1.
  // We accept both domain names and IP addresses.
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = inet_addr(argv[1]);
  if (server.sin_addr.s_addr == -1) {
    // It was not a dotted decimal address - assume it's a domain name,
    // and use gethostbyname() to convert it
    hp = gethostbyname(argv[1]);
    if (hp) {
      server.sin_family = hp->h_addrtype;
      memcpy(&server.sin_addr, hp->h_addr, sizeof(server.sin_addr));
    } 
    else {
      printf("Unknown host %s\n", argv[1]);
      exit(1);
    }
  }

  if (argc >= 3) {
    // Get the port number from command line
    port = atoi(argv[2]);
    if (port > 0 && port < 65535) {
      printf("Using port number %d\n", port);
      port = htons(port);  // change byte order - host to network
    }
    else {
      printf("Incorrect command line. Port number expected\n");
    }
  }
                                            
  if (argc < 2 || port < 1 || port > 65534) {
    // Get the port for the echo protocol out of the etc/services file
    struct servent *echoprot;
    echoprot = getservbyname("echo", "udp");
    if (echoprot == NULL) {
      printf("The echo/udp protocol is not listed in the etc/services file\n");
      exit(1);
    }
    port = echoprot->s_port;
    printf("Using default port number %d\n", ntohs(port));
  }
  server.sin_port = port;

  // Open a UDP socket
  sock = socket(AF_INET, SOCK_DGRAM, 0);
  if (sock == -1) {
    perror("socket()");
    exit(1);
  }

  printf("Enter text line, it will be transmitted to server.\n Enter empty line to exit.\n\n");
  while (1) {
    // Read text line from console
    if (gets(buf) == NULL) {
      printf("gets() returned NULL\n");
      exit(1);
    }

    // If user entered empty string - exit
    if (strlen(buf) == 0) {
      printf("Empty string entered, exiting.\n");
      break;
    }
                    
    // Send line to server
    rc = sendto(sock, buf, strlen(buf), 0, (struct sockaddr *)&server, sizeof(server));
    if (rc == -1) {
      perror("send()");
      exit(1);
    }

    // Receive response from server
    respsize = sizeof(response);
    len = recvfrom(sock, buf, sizeof(buf), 0, (struct sockaddr *)&response, &respsize);
    if (len == -1) {
      perror("recv()");
      exit(1);
    }

    // Print received data to console
    buf[len] = '\0';
    printf("%s\n", buf);
  }
  
  // Close socket and exit
  close(sock);
  return 0;
}
