#include <stdlib.h>
#include <stdio.h>

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <unistd.h>
#include <arpa/inet.h>


int main(int argc, char **argv) {
  struct sockaddr_in server;     // struct for server address
  struct sockaddr_in client;     // struct for client address
  int port;                      // server port number
  int listen_sock;               // listening socket
  int newsock;                   // socket for incoming connection
  char buf[1024];                // buffer for transmit and receive
  int len;                       // length of received data
  int rc;                        // result code
  int asize, i;

  printf("Echo server (single client version)\n"
         "Written by Igor Vanin (vanin@gpmail.spb.ru)\n\n");

  // Create a TCP socket to accept incoming connections
  listen_sock = socket(AF_INET, SOCK_STREAM, 0);
  if (listen_sock == -1) {
    perror("tcp socket()");
    exit(1);
  }

  printf("TCP socket assigned is %d\n", listen_sock);

  if (argc >= 2) {
    // Get the port number from command line
    port = atoi(argv[1]);
    if (port > 0 && port < 65535) {
      printf("Using port number %d\n", port);
      port = htons(port);  // change byte order - host to network
    }
    else {
      printf("Incorrect command line. Port number expected\n");
    }
  }
  
  if (argc < 2 || port < 1 || port > 65534) {
    // Get the port for the echo protocol out of the etc/services file
    struct servent *echoprot;
    echoprot = getservbyname("echo", "tcp");
    if (echoprot == NULL) {
      printf("The echo/tcp protocol is not listed in the etc/services file\n");
      exit(1);
    }
    port = echoprot->s_port;
    printf("Using default port number %d\n", ntohs(port));
  }

  server.sin_family = AF_INET;
  server.sin_port = port;
  server.sin_addr.s_addr = INADDR_ANY;

  // Bind the TCP socket to the echo port
  rc = bind(listen_sock, (struct sockaddr *)&server, sizeof(server));
  if (rc == -1) {
    perror("tcp bind()");
    exit(1);
  }

  // Put the socket into listen mode
  rc = listen(listen_sock, SOMAXCONN);
  if (rc == -1) {
    perror("listen()");
    exit(1);
  }

  while (1) {
    // Wait for a connection on the TCP socket
    asize = sizeof(client);
    newsock = accept(listen_sock, (struct sockaddr *)&client, &asize);
    if (newsock == -1) {
       perror("accept()");
       exit(1);
    }

    printf("Received a TCP connection on socket %d from %s port %d\n", newsock, inet_ntoa(client.sin_addr), ntohs(client.sin_port));
    // Connection established. While working with this client,
    // we don't accept connections from other clients
    while (1) {
      // Read the data
      len = recv(newsock, buf, sizeof(buf), 0);
      if (len == -1) {
        perror("recv()");
        exit(1);
      }

      // The client broke the connection, if recv returned 0
      if (len == 0) {
        printf("The client broke the connection\n");
        close(newsock);
        break;
      }

      printf("Received %d bytes\n", len);

      // Echo it back
      rc = send(newsock, buf, len, 0);
      if (rc == -1) {
        perror("send()");
        exit(1);
      }
    }
  }
}
