#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <unistd.h>
#include <arpa/inet.h>

void Service(int sock);


int main(int argc, char **argv) {
  struct sockaddr_in server;     // struct for server address
  struct sockaddr_in client;     // struct for client address
  int port;                      // server port number
  int listen_sock;               // listening socket
  int newsock;                   // socket for incoming connection
  char buf[1024];                // buffer for transmit and receive
  int rc;                        // result code
  int asize, i;


  printf("Echo server (multi-client fork version)\n"
         "Written by Igor Vanin (vanin@gpmail.spb.ru\n\n");

  // Create a TCP socket to accept incoming connections
  listen_sock = socket(AF_INET, SOCK_STREAM, 0);
  if (listen_sock == -1) {
    perror("tcp socket()");
    exit(1);
  }

  printf("TCP socket assigned is %d\n", listen_sock);


  if (argc >= 2) {
    // Get the port number from command line
    port = atoi(argv[1]);
    if (port > 0 && port < 65535) {
      printf("Using port number %d\n", port);
      port = htons(port);  // change byte order - host to network
    }
    else {
      printf("Incorrect command line. Port number expected\n");
    }
  }
  
  if (argc < 2 || port < 1 || port > 65534) {
    // Get the port for the echo protocol out of the etc/services file
    struct servent *echoprot;
    echoprot = getservbyname("echo", "tcp");
    if (echoprot == NULL) {
      printf("The echo/tcp protocol is not listed in the etc/services file\n");
      exit(1);
    }
    port = echoprot->s_port;
    printf("Using default port number %d\n", ntohs(port));
  }

  server.sin_family = AF_INET;
  server.sin_port = port;
  server.sin_addr.s_addr = INADDR_ANY;

  // Bind the TCP socket to the echo port
  rc = bind(listen_sock, (struct sockaddr *)&server, sizeof(server));
  if (rc == -1) {
    perror("tcp bind()");
    exit(1);
  }

  // Put the socket into listen mode
  rc = listen(listen_sock, SOMAXCONN);
  if (rc == -1) {
    perror("listen()");
    exit(1);
  }

  // Wait for a connection on the TCP socket
  while (1) {
    // Setup the socket array for select()
    fd_set r;
    FD_ZERO(&r);
    FD_SET(listen_sock, &r);

    printf("[%d] Waiting for connection\n", getpid());
    // Wait indefinitely
    rc = select(FD_SETSIZE, &r, NULL, NULL, NULL);
    if (rc == -1) {
      // If the select was interrupted, just continue
      if (errno == EINTR) continue;
      perror("select()");
      exit(1);
    }

    // If this was a request for a new TCP connection, accept it
    if (FD_ISSET(listen_sock, &r)) {
      // accept() the new connection
      asize = sizeof(client);
      newsock = accept(listen_sock, (struct sockaddr *)&client, &asize);
      if (newsock == -1) {
        perror("accept()");
        exit(1);
      }

      printf("[%d] Received a TCP connection on socket %d from %s port %d\n", getpid(), newsock, inet_ntoa(client.sin_addr), ntohs(client.sin_port));
      printf("[%d] Forking...\n", getpid());
      // Calling fork()
      // fork creates the same child process
      // fork returns 0 to new process and child's PID to parent process
      rc = fork();
      if (rc == 0) {
        // New process created
        // Service the connection and exit
        Service(newsock);
        exit(0);
      }
      else if (rc > 0) {
        // This is parent process
        printf("[%d] Created child process with PID=%d\n", getpid(), rc);
      }
      else {
        printf("Error! Can't fork!\n");
        close(newsock);
      }
    }
  }
}


void Service(int sock) {
  int len, rc;
  char buf[1024];

  printf("[%d] This is the child process working with socket %d\n", getpid(), sock);

  while (1) {
    // Read the data
    len = recv(sock, buf, sizeof(buf), 0);
    if (len == -1) {
      perror("recv()");
      exit(1);
    }

    if (len == 0) {
      printf("[%d] The client broke the connection\n", getpid());
      close(sock);
      break;
    }

    printf("[%d] Received %d bytes\n", getpid(), len);

    // Echo it back
    rc = send(sock, buf, len, 0);
    if (rc == -1) {
      perror("send()");
      exit(1);
    }
  }  
}
