#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <unistd.h>
#include <arpa/inet.h>

void Service(int sock);


int main(int argc, char **argv) {
  struct sockaddr_in server;     // struct for server address
  struct sockaddr_in client;     // struct for client address
  int port;                      // server port number
  int udp_sock;                  // UDP socket number
  int rc;                        // result code


  printf("Echo server (UDP)\n"
         "Written by Igor Vanin (vanin@gpmail.spb.ru)\n\n");

  // Create a UDP socket to accept incoming data
  udp_sock = socket(AF_INET, SOCK_DGRAM, 0);
  if (udp_sock == -1) {
    perror("udp socket()");
    exit(1);
  }
  printf("UDP socket assigned is %d\n", udp_sock);

  if (argc >= 2) {
    // Get the port number from command line
    port = atoi(argv[1]);
    if (port > 0 && port < 65535) {
      printf("Using port number %d\n", port);
      port = htons(port);  // change byte order - host to network
    }
    else {
      printf("Incorrect command line. Port number expected\n");
    }
  }
  
  if (argc < 2 || port < 1 || port > 65534) {
    // Get the port for the echo protocol out of the etc/services file
    struct servent *echoprot;
    echoprot = getservbyname("echo", "udp");
    if (echoprot == NULL) {
      printf("The echo/udp protocol is not listed in the etc/services file\n");
      exit(1);
    }
    port = echoprot->s_port;
    printf("Using default port number %d\n", ntohs(port));
  }

  server.sin_family = AF_INET;
  server.sin_port = port;
  server.sin_addr.s_addr = INADDR_ANY;

  // Bind the UDP socket to the echo port
  rc = bind(udp_sock, (struct sockaddr *)&server, sizeof(server));
  if (rc == -1) {
    perror("udp bind()");
    exit(1);
  }

  // Wait for data to echo
  while (1) {
    // Setup the socket array for select()
    fd_set r;
    FD_ZERO(&r);
    FD_SET(udp_sock, &r);

    printf("Waiting for incoming data\n");
    // Wait indefinitely
    rc = select(FD_SETSIZE, &r, NULL, NULL, NULL);
    if (rc == -1) {
      // If the select was interrupted, just continue
      if (errno == EINTR) continue;
      perror("select()");
      exit(1);
    }

    // If we received a UDP packet, handle it
    if (FD_ISSET(udp_sock, &r)) {
      Service(udp_sock);
    }
  }
}


// The UDP socket has a packet waiting to be received
// - receive it and echo it back to the sender
void Service(int sock) {
  static char buf[32767];
  int rc;
  int asize, len;
  struct sockaddr_in client;

  // Receive the datagram
  asize = sizeof(client);
  len = recvfrom(sock, buf, sizeof(buf), 0, (struct sockaddr *)&client, &asize);
  if (len == -1) {
    perror("recvfrom()");
    exit(1);
  }

  printf("Received a UDP echo request from %s port %d\n", inet_ntoa(client.sin_addr), ntohs(client.sin_port));

  // Echo it back
  rc = sendto(sock, buf, len, 0, (struct sockaddr *)&client, sizeof(client));
  if (rc == -1 || rc != len) {
    perror("sendto()");
    exit(1);
  }
}

