#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <sys/time.h>

int in_cksum(u_short *ptr, int len);
int send_ping(int pingsock, struct in_addr addr, long seq);
int recv_ping(int pingsock, struct in_addr addr);


int main(int argc, char **argv) {
  int pingsock;
  struct in_addr addr;
  int packets;
  int sequence;

  printf("Ping utility\n"
         "Written by Igor Vanin (vanin@gpmail.spb.ru)\n\n");

  if (argc < 3) {
    printf("Usage: png <host_address> <number_of_packets>\n");
    return 0;
  }

  addr.s_addr = inet_addr(argv[1]);
  if (addr.s_addr == -1) {
    printf("Incorrect address\n");
    exit(1);
  }

  packets = atoi(argv[2]);

  pingsock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
  if (pingsock == -1) {
    perror("ping socket()");
    exit(1);
  }
  printf("ICMP socket assigned is %d\n", pingsock);

  for (sequence = 0; sequence < packets; sequence++) {
    send_ping(pingsock, addr, sequence);
    recv_ping(pingsock, addr);
  }
  return 0;
}

// send_ping - sends ICMP echo request (type 8, code 0)
int send_ping(int pingsock, struct in_addr addr, long seq) {
  int rc;
  struct timeval tv;
  struct sockaddr_in a;
  char buf[16];
  struct icmp *icmp_pkt;

  icmp_pkt = (struct icmp *)buf;

  memset(&a, 0, sizeof(a));
  a.sin_family = AF_INET;
  a.sin_addr = addr;

  memset(buf, 0, sizeof(buf));
  icmp_pkt->icmp_type = ICMP_ECHO;
  icmp_pkt->icmp_code = 0;
  icmp_pkt->icmp_cksum = 0;
  icmp_pkt->icmp_id   = getpid();
  icmp_pkt->icmp_seq  = seq;
  gettimeofday(&tv, NULL);
  *(unsigned long *)(buf+8) = tv.tv_sec;
  *(unsigned long *)(buf+12) = tv.tv_usec;
  icmp_pkt->icmp_cksum = in_cksum((u_short *)buf, sizeof(buf));

  printf("Sending echo request to %s\n", inet_ntoa(addr));
  rc = sendto(pingsock, (char *)buf, sizeof(buf), 0, (struct sockaddr *)&a, sizeof(a));
  if (rc == -1) {
    perror("sendto");
    return -1;
  }
  return 0;
}


// recv_ping -- receives ICMP echo reply (type 0, code 0)
int recv_ping(int pingsock, struct in_addr addr) {
  int rc;
  char buf[1024];
  struct icmp *icmp_pkt;
  struct sockaddr_in from;
  int fromlen;
  fd_set readmask;
  struct timeval timeout;

  icmp_pkt = (struct icmp *)(buf+20);

  while (1) {
    FD_ZERO(&readmask);
    FD_SET(pingsock, &readmask);
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;
    printf("Waiting...\n");
    rc = select(pingsock+1, &readmask, 0, 0, &timeout);
    switch (rc) { 
      case -1: { // error
        return -1;
      }
      case 0: {
        printf("Timeout, echo reply not received\n");
        return 0;
      }
    }

    fromlen = sizeof(from);
    rc = recvfrom(pingsock, buf, sizeof(buf), 0, (struct sockaddr *)&from, (socklen_t *)&fromlen);
    if (rc == -1) {
      perror("recvfrom");
      return -1;
    }

    if (icmp_pkt->icmp_id == getpid() && (*(long *)&from.sin_addr == *(long *)&addr)) {
      printf("Received %d bytes from %s, seq# %d\n", rc, inet_ntoa(from.sin_addr), icmp_pkt->icmp_seq);
      return 0;
    }
    else {
      printf("Not our packet, ignoring...\n");
    }
  }
}


// in_cksum -- calculates checksum for ICMP packet
int in_cksum(u_short *ptr, int len) {
  int nleft = len;
  u_short *w = ptr;
  int sum = 0;
  u_short answer = 0;

  while (nleft > 1) {
    sum += *w++;
    nleft -= 2;
  }

  if (nleft == 1) {
    *(u_char *)(&answer) = *(u_char *)w;
    sum += answer;
  }

  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);
  answer = ~sum;
  return answer;
}
