#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <sys/time.h>


int main(int argc, char **argv) {
  int icmpsock;
  int rc;
  char buf[1024];
  struct icmp *icmp_pkt;
  struct sockaddr_in from;
  int fromlen;
  fd_set readmask;

  printf("Listen ICMP utility\n"
         "Written by Igor Vanin (vanin@gpmail.spb.ru)\n\n");

  icmpsock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
  if (icmpsock == -1) {
    perror("ICMP socket()");
    exit(1);
  }
  printf("ICMP socket assigned is %d\n", icmpsock);

  icmp_pkt = (struct icmp *)(buf+20);
  while (1) {
    FD_ZERO(&readmask);
    FD_SET(icmpsock, &readmask);
    rc = select(icmpsock+1, &readmask, 0, 0, NULL);
    if (rc == -1) { // error
      exit(1);
    }

    fromlen = sizeof(from);
    rc = recvfrom(icmpsock, buf, sizeof(buf), 0, (struct sockaddr *)&from, (socklen_t *)&fromlen);
    if (rc == -1) {
      perror("recvfrom");
      return -1;
    }

    printf("Received %d bytes from %s. ICMP type %d, code %d\n",
      rc, inet_ntoa(from.sin_addr), icmp_pkt->icmp_type, icmp_pkt->icmp_code);
    switch (icmp_pkt->icmp_type) {
      case 0:
        printf("  echo reply\n");
        break;
      case 3:
        printf("  unreachable ");
        switch (icmp_pkt->icmp_code) {
          case 0:
            printf("(network unreachable)");
            break;
          case 1:
            printf("(host unreachable)");
            break;
          case 2:
            printf("(bad protocol)");
            break;
          case 3:
            printf("(port unreachable)");
            break;
          case 4:
            printf("(fragmentation need but DF bit is set)");
            break;
          case 5:
            printf("(source routing failed)");
            break;
          case 6:
            printf("(unknown net)");
            break;
          case 7:
            printf("(unknown hos)");
            break;
          case 8:
            printf("(source host isolated)");
            break;
          case 9:
            printf("(prohibited access to network)");
            break;
          case 10:
            printf("(prohibited access to host)");
            break;
          case 11:
            printf("(bad tos for ne)");
            break;
          case 12:
            printf("(bad tos for host)");
            break;
        }
        printf("\n");
        break;
      case 4:
        printf("  source quench\n");
        break;
      case 5:
        printf("  redirect ");
        switch (icmp_pkt->icmp_code) {
          case 0:
            printf("(network)");
            break;
          case 1:
            printf("(host)");
            break;
          case 2:
            printf("(tos and net)");
            break;
          case 3:
            printf("(tos and host)");
            break;
        }
        printf("\n");
        break;
      case 8:
        printf("  echo request\n");
        break;
      case 9:
        printf("  router advertisement\n");
        break;
      case 10:
        printf("  router solicitation\n");
        break;
      case 11:
        printf("  time exceeded\n");
        break;
      case 12:
        printf("  ip header bad\n");
        break;
      case 13:
        printf("  timestamp request\n");
        break;
      case 14:
        printf("  timestamp reply\n");
        break;
      case 15:
        printf("  information request\n");
        break;
      case 16:
        printf("  information reply\n");
        break;
      case 17:
        printf("  address mask request\n");
        break;
      case 18:
        printf("  address mask reply\n");
        break;
    }
  }
}

